"""Конфигурация логирования клиентской части"""

import logging
import os
import sys
from client.base.variables import LOGGING_LEVEL

sys.path.append('../')
# Создание объекта-логгера с именем server
_LOGGER = logging.getLogger('client')

# Создание пути и подготовка имени файла для логирования
LOG_PATH = os.path.dirname(os.path.abspath(__file__))
LOG_PATH = os.path.join(LOG_PATH, '../logs/client.log')

# Определен формат сообщений
CLIENT_FORMATTER = logging.Formatter('%(asctime)s %(levelname)s %(filename)s %(message)s')

# Создание файлового обработчика логирования
LOG_FILE = logging.FileHandler(LOG_PATH, encoding='utf8')
LOG_FILE.setFormatter(CLIENT_FORMATTER)
LOG_FILE.setLevel(LOGGING_LEVEL)

# Подключение файлового обработчика к объекту-логгеру
_LOGGER.addHandler(LOG_FILE)
_LOGGER.setLevel(LOGGING_LEVEL)

# для отладки
if __name__ == '__main__':
    _LOGGER.critical('ОТЛАДКА Критическая ошибка')
    _LOGGER.error('ОТЛАДКА Ошибка')
    _LOGGER.debug('ОТЛАДКА Отладочная информация')
    _LOGGER.info('ОТЛАДКА Информационное сообщение')



import dis


class ServerVerifier(type):
    """
    Метакласс, проверяющий что в результирующем классе нет клиентских
    вызовов таких как: connect. Также проверяется, что серверный
    сокет является TCP и работает по IPv4 протоколу.
    """

    # noinspection PyMethodParameters
    def __init__(self, clsname, bases, clsdict):
        # clsname - экземпляр метакласса - Server
        # bases - кортеж базовых классов - ()
        # clsdict - словарь атрибутов и методов экземпляра метакласса

        # Список методов, которые используются в функциях класса
        methods = []
        # Атрибуты, используемые в функциях классов
        attrs = []
        # перебор ключей
        for func in clsdict:
            try:
                # Возвращает итератор по инструкциям в предоставленной функции,
                # методе, строке исходного кода или объекте кода
                ret = dis.get_instructions(clsdict[func])
            # если дескриптор порта
            except TypeError:
                pass
            # если функция разбираем код, получая используемые методы и атрибуты
            else:
                for i in ret:
                    print(i)
                    # i - Instruction(opname='LOAD_GLOBAL', opcode=116, arg=9, argval='send_message',
                    # argrepr='send_message', offset=308, starts_line=201, is_jump_target=False)
                    # opname - имя для операции
                    if i.opname == 'LOAD_GLOBAL':
                        if i.argval not in methods:
                            # заполнение списка методами, использующимися в функциях класса
                            methods.append(i.argval)
                    elif i.opname == 'LOAD_ATTR':
                        if i.argval not in attrs:
                            # заполнение списка атрибутами, использующимися в функциях класса
                            attrs.append(i.argval)
        print(methods)
        # если обнаружено использование недопустимого метода connect, вызывается исключение
        if 'connect' in methods:
            raise TypeError('Использование метода connect недопустимо в серверном классе')
        if not ('SOCK_STREAM' in attrs and 'AF_INET' in attrs):
            raise TypeError('Некорректная инициализация сокета.')
        # Обязательно вызываем конструктор предка
        super().__init__(clsname, bases, clsdict)


class ClientVerifier(type):
    """
    Метакласс, проверяющий что в результирующем классе нет серверных
    вызовов таких как: accept, listen. Также проверяется, что сокет не
    создаётся внутри конструктора класса.
    """

    # noinspection PyMethodParameters
    def __init__(self, clsname, bases, clsdict):
        methods = []
        for func in clsdict:
            try:
                ret = dis.get_instructions(clsdict[func])
            except TypeError:
                pass
            else:
                for i in ret:
                    if i.opname == 'LOAD_GLOBAL':
                        if i.argval not in methods:
                            methods.append(i.argval)
        for command in ('accept', 'listen', 'socket'):
            if command in methods:
                raise TypeError('В классе обнаружено использование запрещённого метода')
        if 'getting_message' in methods or 'sending_message' in methods:
            pass
        else:
            raise TypeError('Отсутствуют вызовы функций, работающих с сокетами.')
        super().__init__(clsname, bases, clsdict)



"""Общие утилиты проекта"""
import json

from client.base.variables import *


def getting_message(sock_obj):
    """
    Утилита приёма и декодирования сообщения.
    Принимает байты преобразует в словарь, если принято что-то другое возвращает ошибку значения
    :param sock_obj:
    :return:
    """
    encoded_response = sock_obj.recv(MAX_PACKAGE_LENGTH)
    if isinstance(encoded_response, bytes):
        response_in_json = encoded_response.decode(ENCODING)
        response = json.loads(response_in_json)
        if isinstance(response, dict):
            return response
        return ValueError
    return ValueError


def sending_message(sock_obj, message):
    """
    Утилита кодирования и отправки сообщения.
    Принимает словарь преобразует в байты
    :param sock_obj:
    :param message:
    :return:
    """
    message_in_json = json.dumps(message)
    encoded_message = message_in_json.encode(ENCODING)
    sock_obj.send(encoded_message)



from setuptools import setup, find_packages

setup(name="mess_client_december",
      version="0.0.1",
      description="mess_client_december",
      author="Alex Morev",
      author_email="iv.iv@yandex.ru",
      packages=find_packages(),
      install_requires=['PyQt5', 'sqlalchemy', 'pycryptodome', 'pycryptodomex']
      )



"""Программа клиента"""

import argparse
import os
import sys

from Crypto.PublicKey import RSA
from PyQt5.QtWidgets import QApplication, QMessageBox

from base.variables import *
import log.config.client_log_config
from client.database import ClientDatabase
from client.main_window import ClientMainWindow
from client.start_dialog import UserNameDialog
from client.transport import ClientTransport


def log_dec(func):
    """Функция-декоратор"""

    def wrap(*args, **kwargs):
        """Обертка"""
        conclusion = func(*args, **kwargs)
        CLIENT_LOGGER.debug(f'Вызвана функция {func.__name__} с параметрами ({args}, {kwargs}). '
                            f'Из модуля {func.__module__}')
        return conclusion

    return wrap


# Инициализация регистратора
CLIENT_LOGGER = logging.getLogger('client')


# noinspection PyShadowingNames
@log_dec
def argument_parser():
    """Парсер параметров коммандной строки,
    чтение параметров, возвращаем 3 параметра.
    """
    parser = argparse.ArgumentParser()
    parser.add_argument('host', default=DEFAULT_IP_ADDRESS, nargs='?')
    parser.add_argument('port', default=DEFAULT_PORT, type=int, nargs='?')
    parser.add_argument('-n', '--name', default=None, nargs='?')
    parser.add_argument('-p', '--password', default='', nargs='?')
    namespace = parser.parse_args(sys.argv[1:])
    server_address = namespace.host
    server_port = namespace.port
    client_name = namespace.name
    client_passwd = namespace.password

    # проверка корректности номера порта
    if not 1023 < server_port < 65536:
        CLIENT_LOGGER.critical(f'Попытка запуска клиента с указанием некорректного порта {server_port}')
        sys.exit(1)

    return server_address, server_port, client_name, client_passwd


if __name__ == '__main__':
    """
    Загрузка параметров коммандной строки
    """
    server_address, server_port, client_name, client_passwd = argument_parser()

    # создание клиентского приложения
    client_app = QApplication(sys.argv)

    # Если пользователь не был указан в командной строке, запрашивается его имя
    start_dialog = UserNameDialog()
    if not client_name or not client_passwd:
        client_app.exec_()
        # Если пользователь ввёл имя и нажал ОК, то сохраняем ведённое и удаляем объект, инааче выходим
        if start_dialog.ok_pressed:
            client_name = start_dialog.client_name.text()
            client_passwd = start_dialog.client_passwd.text()
            CLIENT_LOGGER.debug(f'Using USERNAME = {client_name}, PASSWD = {client_passwd}.')
        else:
            sys.exit(0)

    CLIENT_LOGGER.info(f'Клиент запущен - адрес хоста: {server_address} порт: {server_port}, '
                       f'пользователь: {client_name}')

    # Загружаем ключи с файла, если же файла нет, то генерируем новую пару.
    dir_path = os.getcwd()
    key_file = os.path.join(dir_path, f'{client_name}.key')
    if not os.path.exists(key_file):
        keys = RSA.generate(2048, os.urandom)
        with open(key_file, 'wb') as key:
            key.write(keys.export_key())
    else:
        with open(key_file, 'rb') as key:
            keys = RSA.import_key(key.read())

    # !!!keys.publickey().export_key()
    CLIENT_LOGGER.debug("Keys successfully loaded.")

    # Инициализация БД
    database = ClientDatabase(client_name)

    # Инициализация сокета и запуск потока
    try:
        tcp_sock = ClientTransport(server_port, server_address, database, client_name, client_passwd, keys)
        CLIENT_LOGGER.debug("Socket ready.")
    except Exception as ErrorNotConnection:
        message = QMessageBox()
        message.critical(start_dialog, 'Ошибка сервера', f'{ErrorNotConnection} - нет соединения')
        CLIENT_LOGGER.error(f'Не удалось подключиться к серверу {server_address}:{server_port}')
        sys.exit(1)
    tcp_sock.setDaemon(True)
    tcp_sock.start()

    # Создание GUI
    main_window = ClientMainWindow(database, tcp_sock, keys)
    main_window.make_connection(tcp_sock)
    main_window.setWindowTitle(f'Чат-программа b-release - {client_name}')
    client_app.exec_()

    # После закрытия графической оболочки, закрываем сокет
    tcp_sock.transport_shutdown()
    tcp_sock.join()
